import os
import sys
import json
import click
import pandas
import tables
import random
import warnings
import numpy as np

#from hydra_pywr import *
from pywr.model import Model
from .moea import SaveNondominatedSolutionsArchive
from pywr.recorders.progress import ProgressRecorder
from .custom_parameters_and_recorders import *
from .preprocess import *

import logging
logger = logging.getLogger(__name__)

@click.group()
def cli():
    pass

@cli.command(name='run', help="Run a Pywr model from a JSON file.")
@click.option('-f', '--file-name', type=str, help="The JSON file of the Pywr model you would like to run.")
def run(file_name):
    path = os.path.join("outputs")
    os.makedirs(path, exist_ok=True)

    #preprocessjson = JSONPreprocess(json_file=file_name)
    #preprocessjson.reorder()

    """ Run the Pywr model. """
    logger.info('Loading model from file: "{}"'.format(file_name))
    model = Model.load(file_name)

    warnings.filterwarnings('ignore', category=tables.NaturalNameWarning)

    ProgressRecorder(model)

    logger.info('Starting model run.')
    ret = model.run()
    logger.info(ret)

    try:
        df = model.to_dataframe()

        fn = '{}/{}.csv'.format("outputs","water_results_time_step")
        i=1
        while os.path.isfile(os.path.join(os.getcwd(), fn)):
            fn = '{}/{}_{}.csv'.format("outputs","water_results_time_step", str(i))
            i+=1
        df.to_csv(fn)
    except:
        pass


@cli.command(name='search', help="Perform multi-objective optimization for a Pywr JSON file model using an MOEA algorithm.")
@click.option('-f', '--file-name', type=str,help="The JSON file of the pywr model you would like to do the search on.")
@click.option('-s', '--seed', type=int, default=None, help="An integer representing the random starting point for the search.")
@click.option('-p', '--num-cpus', type=int, default=None, help="The number of processing cores to used for the search.")
@click.option('-n', '--max-nfe', type=int, default=1000, help="The number of iterations for the search.")
@click.option('--pop-size', type=int, default=50, help="The population size of the search.")
@click.option('-a', '--algorithm', type=click.Choice(['NSGAII', 'NSGAIII', 'EpsMOEA', 'EpsNSGAII']), default='NSGAII', help="The MOEA algorithm to be used for the search.")
@click.option('-e', '--epsilons', multiple=True, type=float, default=(0.05, ), help="The Epsilon value used with the EpsMOEA and EpsNSGAII algorithms.")
@click.option('--divisions-outer', type=int, default=12, help="The number of outer divisions used with the NSGAIII algorithm.")
@click.option('--divisions-inner', type=int, default=0, help="The number of inner divisions used with the NSGAIII algorithm.")
def search(file_name, seed, num_cpus, max_nfe, pop_size, algorithm, epsilons, divisions_outer, divisions_inner):
    import platypus

    logger.info('Loading model from file: "{}"'.format(file_name))
    directory, model_name = os.path.split(file_name)
    output_directory = os.path.join(directory, 'outputs')

    #preprocessjson = JSONPreprocess(json_file=file_name)
    #preprocessjson.reorder()

    if algorithm == 'NSGAII':
        algorithm_klass = platypus.NSGAII
        algorithm_kwargs = {'population_size': pop_size}
    elif algorithm == 'NSGAIII':
        algorithm_klass = platypus.NSGAIII
        algorithm_kwargs = {'divisions_outer': divisions_outer, 'divisions_inner': divisions_inner}
    elif algorithm == 'EpsMOEA':
        algorithm_klass = platypus.EpsMOEA
        algorithm_kwargs = {'population_size': pop_size, 'epsilons': epsilons}
    elif algorithm == 'EpsNSGAII':
        algorithm_klass = platypus.EpsNSGAII
        algorithm_kwargs = {'population_size': pop_size, 'epsilons': epsilons}
    else:
        raise RuntimeError('Algorithm "{}" not supported.'.format(algorithm))

    if seed is None:
        seed = random.randrange(sys.maxsize)

    search_data = {'algorithm': algorithm, 'seed': seed, 'user_metadata':algorithm_kwargs}
    wrapper = SaveNondominatedSolutionsArchive(file_name, search_data=search_data, output_directory=output_directory,
                                                   model_name=model_name)

    if seed is not None:
        random.seed(seed)

    logger.info('Starting model search.')

    if num_cpus is None:
        evaluator_klass = platypus.MapEvaluator
        evaluator_args = ()
    else:
        evaluator_klass = platypus.ProcessPoolEvaluator
        evaluator_args = (num_cpus,)

    with evaluator_klass(*evaluator_args) as evaluator:
        algorithm = algorithm_klass(wrapper.problem, evaluator=evaluator, **algorithm_kwargs, seed=seed)

        algorithm.run(max_nfe, callback=wrapper.save_nondominant)

def start_cli():
    """ Start the command line interface. """
    from . import logger
    import sys
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler(stream=sys.stdout)
    logger.addHandler(ch)

    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    # Also log pywr messages
    pywr_logger = logging.getLogger('pywr')
    pywr_logger.setLevel(logging.INFO)
    pywr_logger.addHandler(ch)
    cli(obj={})

start_cli()



import uuid
import gzip
import os
import datetime
import pandas as pd
import numpy as np
import platypus
from platypus import Hypervolume
#from hydra_pywr import *

from pywr.optimisation.platypus import PlatypusWrapper
from .custom_parameters_and_recorders import *


import logging
logger = logging.getLogger(__name__)

class SaveNondominatedSolutionsArchive(PlatypusWrapper):

    def __init__(self, *args, **kwargs):
        self.output_directory = kwargs.pop('output_directory', 'outputs')
        self.search_data = kwargs.pop('search_data', {})
        self.model_name = kwargs.pop('model_name')
        super().__init__(*args, **kwargs)

        # To determine the number of variables, etc
        m = self.model
        self.hv_calculated = []

    @property
    def output_subdirectory(self):
        path = os.path.join(self.output_directory, self.model_name[0:-4], f'{self.uid}_seed{self.search_data["seed"]}')
        os.makedirs(path, exist_ok=True)
        return path

    def save_nondominant(self, algorithm):

        uid = uuid.uuid4().hex

        va = []
        va_ = []
        nva_ = []

        platypus_variables = platypus.nondominated(algorithm.result)
        for v in platypus_variables:
            for ivar, var in enumerate(self.model_variables):
                j = slice(self.model_variable_map[ivar], self.model_variable_map[ivar+1])
                tem_va = np.array(v.variables[j])
                tem_nme = [f'{var.name}[d{i}]' for i in range(len(tem_va))]

                va_.append(tem_va)
                nva_.append(tem_nme)

            va_flatt = [val for sublist in va_ for val in sublist]
            nva_flatt = [val for sublist in nva_ for val in sublist]
            va.append(va_flatt)
            va_ = []
            nva_ = []

        variables = pd.DataFrame(data=va, columns=nva_flatt)
        variables = variables.add_prefix('VAR_')

        platypus_objectives = platypus.nondominated(algorithm.result)

        objectives = pd.DataFrame(data=np.array([o.objectives for o in platypus_objectives]),
                                  columns=[f'{o.name}' for o in self.model_objectives])

        objectives = objectives.add_prefix('OBJ_')

        platypus_constraints = platypus.nondominated(algorithm.result)

#        TODO allow to pass double bounded constraints
        constraints = pd.DataFrame(data=np.array([c.constraints for c in platypus_constraints]),
                                   columns=[f'{c.name}' for c in self.model_constraints])

        constraints = constraints.add_prefix('CONS_')

        metrics = pd.concat([objectives, constraints, variables], axis=1)
        metrics['NFE'] = algorithm.nfe
        metrics.to_csv(os.path.join(self.output_subdirectory, f'nfe{algorithm.nfe}.csv.gz'), compression='infer')


        # Compute hypervolume
        num_objectives = len(self.model_objectives)
        min_ref = []
        max_ref = []
        for i in range(num_objectives):
                min_ref.append(-20000000)
                max_ref.append(20000000)

        hyp = Hypervolume(minimum=min_ref, maximum=max_ref)
        hv = hyp.calculate(algorithm.result)
        
        self.hv_calculated.append([algorithm.nfe, hv])
        hv_pd = pd.DataFrame(data=self.hv_calculated, columns=['nfe','hypervolume']).set_index('nfe')
        hv_pd.to_csv(os.path.join(self.output_subdirectory, f'hypervolume_seed_{self.search_data["seed"]}.csv'))
        print("Hypervolume:", hv)



from setuptools import setup, find_packages

with open('README.rst') as f:
    long_description = f.read()

setup(
    name='wrh',
    version='0.9',
    long_description=long_description,
    long_description_content_type="text/x-rst",
    description='Package for CIV550 course and UofT.',
    url='https://wrh.civmin.utoronto.ca/',
    author='Mohammed Basheer',
    author_email='mohammedadamabbaker@gmail.com',
    install_requires=['click','pandas','numpy','scipy','platypus-opt','tables'],
    packages=find_packages(),
    package_data={
        'wrh': ['json/*.json'],
    },
    entry_points={
        'console_scripts': ['wrh=wrh.cli:start_cli'],
    },
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Operating System :: OS Independent',
        'Natural Language :: English'
    ]
)
